#
# Copyright (c) 2013 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

__all__ = [
    'application'
]

from wsgiref.simple_server import make_server
from webob import exc
from jinja2 import Environment, FileSystemLoader
from wtforms import Form
from wtforms.fields import TextField, PasswordField
from wtforms.validators import (Optional, Required, EqualTo, Regexp,
                                ValidationError)
from yubiauth.config import settings
from yubiauth.util import validate_otp, MODHEX
from yubiauth.util.rest import REST_API, Route, extract_params
from yubiauth.client.rest import session_api, require_session
from yubiauth.client.controller import requires_otp

import os
import logging
log = logging.getLogger(__name__)

base_dir = os.path.dirname(__file__)
template_dir = os.path.join(base_dir, 'templates')
env = Environment(loader=FileSystemLoader(template_dir))

YUBIKEY_OTP = r'[%s]{32,64}' % MODHEX
YUBIKEY_PREFIX = r'[%s]{2,32}' % MODHEX


def redirect(request, target):
    return exc.HTTPSeeOther(location=request.relative_url(target, True))


def with_yubikey(func):
    def inner(self, request, prefix, *args, **kwargs):
        user = request.environ['yubiauth.user']
        if not prefix in user.yubikeys:
            raise exc.HTTPForbidden(detail='Unauthorized')
        return func(self, request, user.yubikeys[prefix], *args, **kwargs)
    return inner


class LoginForm(Form):
    username = TextField('Username', [Optional() if settings['yubikey_id']
                                      else Required()])
    password = PasswordField('Password', [Optional() if
                                          settings['allow_empty'] else
                                          Required()])
    yubikey = TextField('YubiKey', [Optional(), Regexp(YUBIKEY_OTP)])


class RegisterForm(Form):
    username = TextField('Username', [Required()])
    password = PasswordField('Password', [Required()])
    verify_password = PasswordField(
        'Repeat password',
        [Required(), EqualTo('password', 'Passwords do not match!')]
    )
    yubikey = TextField('YubiKey', [Optional(), Regexp(YUBIKEY_OTP)])


class ReauthenticateForm(Form):
    legend = "Re-authenticate"
    description = "Please re-authenticate to complete this action"
    password = PasswordField('Password')
    otp = TextField('YubiKey OTP', [Optional(), Regexp(YUBIKEY_OTP)])

    def __init__(self, request, prefix=None):
        super(ReauthenticateForm, self).__init__(request.params)
        user = request.environ['yubiauth.user']
        self.username = user.name
        self.client = request.environ['yubiauth.client']
        self.prefix = prefix
        if self.prefix:
            self.otp.description = 'OTP from YubiKey with prefix: %s' % prefix
        elif not requires_otp(user):
            del self.otp

    def validate_otp(self, field):
        if self.prefix and self.prefix != field.data[:-32]:
            raise ValidationError('OTP must have prefix %s' % self.prefix)

    def authenticate(self):
        password = self.password.data
        otp = self.data.get('otp', None)
        try:
            self.client.authenticate(self.username, password, otp)
            return True
        except:
            pass
        return False


class AssignYubikeyForm(Form):
    legend = "Assign new YubiKey"
    yubikey = TextField('New Yubikey OTP', [Regexp(YUBIKEY_OTP)])


class ChangePasswordForm(Form):
    legend = "Change password"
    new_password = PasswordField('New Password', [Required()])
    verify_password = PasswordField(
        'Repeat password',
        [Required(), EqualTo('new_password', 'Passwords do not match!')]
    )


class ClientUI(REST_API):
    __yubikey__ = r'/yubikey/(%s)' % YUBIKEY_PREFIX

    __routes__ = [
        Route(r'^/$', 'index'),
        Route(r'^/revoke$', 'revoke'),
        Route(r'^/login$', post='login'),
        Route(r'^/register$', post='register'),
        Route(r'^/logout$', 'logout'),
        Route(r'^/status$', 'status'),
        Route(r'^/assign_yubikey$', post='assign_yubikey'),
        Route(r'^/change_password$', 'change_password'),
        Route(r'^/delete_account$', 'delete_account'),
        Route(__yubikey__ + r'$', 'yubikey'),
        Route(__yubikey__ + r'/enable$', 'yubikey_enable'),
        Route(__yubikey__ + r'/disable$', 'yubikey_disable'),
        Route(__yubikey__ + r'/generate$', 'yubikey_generate'),
        Route(__yubikey__ + r'/unassign$', 'yubikey_unassign'),
    ]

    def add_message(self, message, level=None):
        self._messages.append({'text': message, 'level': level})

    def _call_setup(self, request):
        super(ClientUI, self)._call_setup(request)
        self._messages = []

    def session_required(self, request, e):
        return self.render(request, 'session_required')

    def render(self, request, tmpl, **data):
        template = env.get_template('%s.html' % tmpl)
        data['base_url'] = '%s/' % request.script_name
        data['messages'] = self._messages
        if 'yubiauth.user' in request.environ:
            data['user'] = request.environ['yubiauth.user']
        return template.render(data)

    def index(self, request, login_form=LoginForm(),
              register_form=RegisterForm()):
        login_form.yubikey.data = None
        if settings['registration']:
            return self.render(request, 'register', login_form=login_form,
                               register_form=register_form)
        else:
            return self.render(request, 'login', login_form=login_form)

    def revoke(self, request):
        if 'revoke' in request.params:
            client = request.environ['yubiauth.client']
            try:
                client.revoke(request.params['revoke'])
                self.add_message('YubiKey revoked!', 'success')
            except:
                self.add_message('Invalid revocation code!', 'error')
        return self.render(request, 'revoke')

    def register(self, request):
        register_form = RegisterForm(request.params)
        if register_form.validate():
            client = request.environ['yubiauth.client']
            username = register_form.username.data
            password = register_form.password.data
            otp = register_form.yubikey.data
            if not otp:
                otp = None
            try:
                user = client.register(username, password, otp)
                return self.render(request, 'created', user=user,
                                   login_form=LoginForm())
            except Exception, e:
                self.add_message('Account registration failed!', 'error')
                log.warn(e)
        return self.index(request, register_form=register_form)

    @extract_params('username?', 'password?', 'yubikey?')
    def login(self, request, username=None, password=None, yubikey=None):
        login_form = LoginForm(request.params)
        if login_form.validate():
            client = request.environ['yubiauth.client']
            try:
                session = client.create_session(username, password, yubikey)
                request.environ['beaker.session'].update(session)
                session.delete()
                return redirect(request, 'status')
            except Exception:
                self.add_message('Login failed!', 'error')
                request.environ['beaker.session'].delete()
        return self.index(request, login_form=login_form)

    @require_session
    def status(self, request):
        return self.render(request, 'status')

    @require_session
    @extract_params('noauth?')
    def assign_yubikey(self, request, noauth=None):
        assign_form = AssignYubikeyForm(request.params)
        auth_form = ReauthenticateForm(request)
        user = request.environ['yubiauth.user']
        if noauth is not None or not assign_form.validate():
            pass
        elif auth_form.validate() and auth_form.authenticate():
            yubikey = assign_form.yubikey.data
            prefix = yubikey[:-32]
            if not validate_otp(yubikey):
                self.add_message('Invalid OTP for new YubiKey!', 'error')
            if not prefix in user.yubikeys:
                user.assign_yubikey(prefix)
            return redirect(request, 'status')
        else:
            self.add_message('Invalid credentials!', 'error')

        return self.render(request, 'assign_yubikey',
                           fieldsets=[assign_form, auth_form])

    @require_session
    @with_yubikey
    def yubikey_unassign(self, request, yubikey):
        auth_form = ReauthenticateForm(request)
        if request.method == 'POST' and auth_form.validate():
            if auth_form.authenticate():
                user = request.environ['yubiauth.user']
                del user.yubikeys[yubikey.prefix]
                return redirect(request, 'status')
            else:
                self.add_message('Invalid credentials!', 'error')

        return self.render(request, 'reauthenticate', yubikey=yubikey,
                           fieldsets=[auth_form], target=request.path_info[1:])

    @require_session
    def change_password(self, request):
        password_form = ChangePasswordForm(request.params)
        auth_form = ReauthenticateForm(request)
        if request.method == 'POST' and password_form.validate() and \
                auth_form.validate():
            if auth_form.authenticate():
                new_password = password_form.new_password.data
                user = request.environ['yubiauth.user']
                user.set_password(new_password)
                return redirect(request, 'status')
            else:
                self.add_message('Invalid credentials!', 'error')

        return self.render(request, 'change_password',
                           fieldsets=[password_form, auth_form])

    @require_session
    def delete_account(self, request):
        auth_form = ReauthenticateForm(request)
        if request.method == 'POST' and auth_form.validate():
            if auth_form.authenticate():
                user = request.environ['yubiauth.user']
                user.delete()
                return redirect(request, '')
            else:
                self.add_message('Invalid credentials!', 'error')

        return self.render(request, 'reauthenticate',
                           fieldsets=[auth_form])

    @require_session(error_handler=lambda req, *x: redirect(req, ''))
    def logout(self, request):
        request.environ['beaker.session'].delete()
        return redirect(request, '')

    @require_session
    @with_yubikey
    def yubikey(self, request, yubikey):
        return self.render(request, 'yubikey', yubikey=yubikey)

    @require_session
    @with_yubikey
    def yubikey_set_enabled(self, request, yubikey, enabled):
        auth_form = ReauthenticateForm(request, yubikey.prefix)
        if request.method == 'POST' and auth_form.validate():
            user = request.environ['yubiauth.user']
            # Validate otp manually as the YubiKey might be disabled
            if user.validate_password(auth_form.password.data) and \
                    validate_otp(auth_form.otp.data):
                yubikey.enabled = enabled
                return redirect(request, 'yubikey/%s' % yubikey.prefix)
            else:
                self.add_message('Invalid credentials!', 'error')

        return self.render(request, 'reauthenticate', yubikey=yubikey,
                           fieldsets=[auth_form], target=request.path_info[1:])

    def yubikey_enable(self, *args, **kwargs):
        return self.yubikey_set_enabled(*args, enabled=True, **kwargs)

    def yubikey_disable(self, *args, **kwargs):
        return self.yubikey_set_enabled(*args, enabled=False, **kwargs)

    @require_session
    @with_yubikey
    def yubikey_generate(self, request, yubikey):
        auth_form = ReauthenticateForm(request, yubikey.prefix)
        if request.method == 'POST' and auth_form.validate():
            if auth_form.authenticate():
                client = request.environ['yubiauth.client']
                code = client.generate_revocation(yubikey.prefix)
                return self.render(request, 'revocation_code', yubikey=yubikey,
                                   code=code)
            else:
                self.add_message('Invalid credentials!', 'error')
        return self.render(request, 'reauthenticate', yubikey=yubikey,
                           fieldsets=[auth_form], target=request.path_info[1:])

application = session_api(ClientUI())

if __name__ == '__main__':
    httpd = make_server('localhost', 8080, application)
    httpd.serve_forever()



#
# Copyright (c) 2013 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

__all__ = [
    'session_api',
    'require_session',
    'ClientAPI',
    'application',
]

from wsgiref.simple_server import make_server
from webob import exc
from webob.dec import wsgify
from beaker.middleware import SessionMiddleware
from yubiauth.client import Client
from yubiauth.util import validate_otp
from yubiauth.util.rest import (REST_API, Route, json_response, json_error,
                                extract_params)
from yubiauth import settings

import logging
log = logging.getLogger(__name__)

SESSION_COOKIE = 'YubiAuth-Session'
SESSION_HEADER = 'X-%s' % SESSION_COOKIE
REVOKE_ATTRIBUTE = '_REVOKE'


@wsgify.middleware
def ClientMiddleware(request, app):
    # Allow the session ID to be provided as a header or cookie.
    # Header takes precedence over cookie.
    if SESSION_HEADER in request.headers:
        request.cookies[SESSION_COOKIE] = request.headers[SESSION_HEADER]
    if 'yubiauth.client' in request.environ:
        return app
    try:
        with Client() as client:
            request.environ['yubiauth.client'] = client
            response = request.get_response(app)
    finally:
        del request.environ['yubiauth.client']
    return response


def session_api(app):
    return ClientMiddleware(SessionMiddleware(app, settings['beaker']))


def require_session(func=None, **kwargs):
    """
    Used to decorate a method on a REST_API to ensure that the user has
    a valid UserSession when entering the method. If not, an error will be
    returned.

    To customize the error, override the session_required method of the api,
    or pass an explicit error handler to the decorator.

    This decorator expects a valid beaker.session and yubiauth.client in the
    environ. In turn, it provides a User object as yubiauth.user.
    """

    error_handler = kwargs.get('error_handler', None)

    def inner(func):
        def new_func(self, request, *args, **kwargs):
            try:
                session = request.environ['beaker.session']
                client = request.environ['yubiauth.client']
                user_id = session.get('user_id', None)
                request.environ['yubiauth.user'] = \
                    client.auth.get_user(user_id)
            except Exception, e:
                log.exception('Session required!')
                if error_handler:
                    return error_handler(request, e)
                elif hasattr(self, 'session_required'):
                    return self.session_required(request, e)
                else:
                    raise exc.HTTPBadRequest(detail='Session required!')
            return func(self, request, *args, **kwargs)
        return new_func
    # If func is not defined, the decorator was called with parentheses.
    return inner(func) if func else inner


class ClientAPI(REST_API):
    __routes__ = [
        Route(r'^/login$', 'login'),
        Route(r'^/authenticate$', 'authenticate'),
        Route(r'^/logout$', 'logout'),
        Route(r'^/status$', 'status'),
        Route(r'^/password$', post='change_password'),
        Route(r'^/yubikey$', post='assign_yubikey'),
        Route(r'^/revoke/generate$', post='generate_revocation'),
        Route(r'^/revoke$', post='revoke_yubikey')
    ]

    def session_required(self, request, e):
        return json_error('Session required!')

    @extract_params('username?', 'password?', 'otp?')
    def authenticate(self, request, username=None, password=None, otp=None):
        try:
            client = request.environ['yubiauth.client']
            client.authenticate(username, password, otp)
            return json_response(True)
        except:
            return json_response(False, status=400)

    @extract_params('username?', 'password?', 'otp?')
    def login(self, request, username=None, password=None, otp=None):
        client = request.environ['yubiauth.client']
        try:
            session = client.create_session(username, password, otp)
            request.environ['beaker.session'].update(session)
            session.delete()
            return json_response(True)
        except Exception, e:
            log.warn(e)
            return json_error('Invalid credentials!')

    @require_session
    def logout(self, request):
        request.environ['beaker.session'].delete()
        return json_response(True)

    @require_session
    def status(self, request):
        return json_response(request.environ['beaker.session']._session())

    @require_session
    @extract_params('oldpass', 'newpass', 'otp?')
    def change_password(self, request, oldpass, newpass, otp=None):
        client = request.environ['yubiauth.client']
        user = request.environ['yubiauth.user']
        try:
            client.authenticate(user.name, oldpass, otp)
            user.set_password(newpass)
            return json_response(True)
        except:
            return json_error('Invalid credentials!')

    @require_session
    @extract_params('yubikey', 'password', 'otp?')
    def assign_yubikey(self, request, yubikey, password, otp=None):
        client = request.environ['yubiauth.client']
        user = request.environ['yubiauth.user']
        try:
            client.authenticate(user.name, password, otp)
            prefix = yubikey[:-32]
            if not validate_otp(yubikey):
                return json_error('Invalid OTP for new YubiKey!')
            if not prefix in user.yubikeys:
                user.assign_yubikey(prefix)
            return json_response(True)
        except:
            return json_error('Invalid credentials!')

    @require_session
    @extract_params('yubikey', 'password', 'otp?')
    def generate_revocation(self, request, yubikey, password, otp=None):
        client = request.environ['yubiauth.client']
        user = request.environ['yubiauth.user']
        try:
            client.authenticate(user.name, password, otp)
            if not user.validate_otp(otp):
                return json_error('Invalid credentials!')
            code = client.generate_revocation(otp[:-32])
            return json_response(code)
        except:
            return json_error('Invalid credentials!')

    @extract_params('code')
    def revoke_yubikey(self, request, code):
        client = request.environ['yubiauth.client']
        try:
            client.revoke(code)
            return json_response(True)
        except:
            return json_error('Invalid code!')


application = session_api(ClientAPI())

if __name__ == '__main__':
    httpd = make_server('localhost', 8080, application)
    httpd.serve_forever()



# Copyright (c) 2013 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

__all__ = [
    'Route',
    'REST_API',
    'json_response',
    'json_error',
    'extract_params'
]

from webob import exc, Response, Request
from webob.dec import wsgify

import json
import re


def no_content():
    return exc.HTTPNoContent()
    #return Response(status=204, content_type=None)


def json_response(data, **kwargs):
    return Response(json.dumps(data), content_type='application/json',
                    **kwargs)


def json_error(message, **kwargs):
    if not 'status' in kwargs:
        kwargs['status'] = 400
    return json_response({'error': message}, **kwargs)


class extract_params(object):
    """
    Decorator for extracting request parameters into kwargs.
    Suffix the parameter with a question mark (?) to make it optional.
    """
    def __init__(self, *params):
        self.params = params

    def _find_request(self, args):
            for arg in args:
                if type(arg) == Request:
                    self.request = arg
                    return True
            return False

    def _extract_params(self):
        self.extracted = {}
        for param in self.params:
            if param.endswith('?'):
                param = param[:-1]
            elif not param in self.request.params:
                return False
            if param in self.request.params and not param in self.extracted:
                self.extracted[param] = self.request.params[param]
        return True

    def __call__(self, func):
        def inner(*args, **kwargs):
            if not self._find_request(args):
                return json_error('Unable to find request!', status=500)

            if not self._extract_params():
                return json_error('Missing required parameter(s)!')

            kwargs.update(self.extracted)
            return func(*args, **kwargs)
        return inner


class Route(object):
    def __init__(self, pattern_str, controller=None, **kwargs):
        self.pattern = re.compile(pattern_str)

        if controller:
            self.get = controller
            self.post = controller
        if 'get' in kwargs:
            self.get = kwargs['get']
        if 'post' in kwargs:
            self.post = kwargs['post']
        if 'delete' in kwargs:
            self.delete = kwargs['delete']

    def get_controller(self, request):
        match = self.pattern.match(request.path_info)

        if match:
            try:
                controller = self.__getattribute__(request.method.lower())
                return controller, match.groups()
            except AttributeError:
                return json_error('Method %s not allowed' % request.method,
                                  status=405), None

        return None, None


class REST_API(object):
    __routes__ = []

    @wsgify
    def __call__(self, request, *args, **kwargs):
        for route in self.__routes__:
            controller, sub_args = route.get_controller(request)
            if controller:
                if isinstance(controller, Response):
                    return controller
                self._call_setup(request)
                args += sub_args
                response = None
                try:
                    controller = self.__getattribute__(controller)
                    response = controller(request, *args, **kwargs)
                finally:
                    self._call_teardown(request, response)
                return response

        raise exc.HTTPNotFound

    def _call_setup(self, request):
        pass

    def _call_teardown(self, request, response):
        pass



#
# Copyright (c) 2013 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

__all__ = [
    'User',
    'Attribute',
    'YubiKey',
    'Base',
]

from UserDict import DictMixin
from yubiauth.config import settings
from yubiauth.util import validate_otp
from yubiauth.util.model import Session, Deletable

from sqlalchemy import (Sequence, Column, Boolean, Integer, String, ForeignKey,
                        UniqueConstraint, Table)
from sqlalchemy.ext.declarative import declarative_base, declared_attr
from sqlalchemy.orm import relationship, backref
from sqlalchemy.orm.collections import attribute_mapped_collection
from sqlalchemy.ext.associationproxy import association_proxy

from passlib.context import CryptContext
from passlib.registry import register_crypt_handler_path

if settings['use_hsm']:
    register_crypt_handler_path('yhsm_pbkdf2_sha1', 'yubiauth.yhsm')
    register_crypt_handler_path('yhsm_pbkdf2_sha256', 'yubiauth.yhsm')
    register_crypt_handler_path('yhsm_pbkdf2_sha512', 'yubiauth.yhsm')


Base = declarative_base()

pwd_context = CryptContext(**settings['crypt_context'])


user_yubikeys = Table(
    'user_yubikeys',
    Base.metadata,
    Column('user_id', Integer, ForeignKey('users.id', ondelete='cascade')),
    Column('yubikey_id', Integer,
           ForeignKey('yubikeys.id', ondelete='cascade'))
)


class AttributeAssociation(Base):
    """
    Holds Attributes. Each AttributeHolder has one of these.

    cvar id: Primary key.
    cvar discriminator: The name of the type of the owner.
    cvar attributes: Dict of attributes.
    cvar owner: The owner of the attributes.
    """
    __tablename__ = 'attribute_associations'

    id = Column(Integer, Sequence('attr_assoc_seq'), primary_key=True)
    discriminator = Column('discriminator', String(16))
    _attributes = relationship(
        'Attribute',
        backref='association',
        order_by='Attribute.key',
        collection_class=attribute_mapped_collection('key'),
        cascade='all, delete-orphan'
    )
    attributes = association_proxy(
        '_attributes',
        'value',
        creator=lambda k, v: Attribute(k, v)
    )

    @property
    def owner(self):
        return getattr(self, '%s_owner' % self.discriminator)


class Attribute(Base):
    """
    Holds an attribute for an AttributeHolder.

    An AttributeHolder can have zero or more attributes, though each attribute
    must have a unique key per user.

    cvar id: Primary key.
    cvar key: The key of the attribute.
    cvar value: The value of the attribute.
    cvar owner: The owner of the attribute.
    """
    __tablename__ = 'attributes'
    __table_args__ = (UniqueConstraint('association_id', 'key',
                                       name='_owner_key_uc'),)

    id = Column(Integer, Sequence('attribute_id_seq'), primary_key=True)
    _association_id = Column('association_id', Integer, ForeignKey(
        'attribute_associations.id'))
    key = Column(String(32), nullable=False)
    value = Column(String(128), nullable=False)
    owner = association_proxy('association', 'owner')

    def __init__(self, key, value):
        self.key = key
        self.value = value

    def __repr__(self):
        return ('%s = %s' % (self.key, self.value)).encode('utf-8')


class AttributeProxy(DictMixin):
        """
        Proxy used in AttributeHolder used to give access to Attributes
        via an AttributeAssociation object, creating that object when
        necessary.
        """
        def __init__(self, owner):
            self.owner = owner

        @property
        def has_assoc(self):
            return bool(self.owner._attribute_association)

        @property
        def assoc(self):
            if not self.owner._attribute_association:
                discriminator = self.owner.__class__.__name__.lower()
                self.owner._attribute_association = AttributeAssociation(
                    discriminator=discriminator)
            return self.owner._attribute_association

        def __getitem__(self, key):
            if self.has_assoc:
                return self.assoc.attributes[key]
            raise KeyError(key)

        def __setitem__(self, key, value):
            self.assoc.attributes[key] = value

        def __delitem__(self, key):
            if self.has_assoc:
                del self.assoc.attributes[key]
            else:
                raise KeyError(key)

        def keys(self):
            if self.has_assoc:
                return self.assoc.attributes.keys()
            return []

        def copy(self):
            copy = {}
            copy.update(self)
            return copy


class AttributeHolder(object):
    """
    Mixin class for model objects that should hold Attributes.

    Attributes can be easily accessed through common dict operations.

    holder.attributes['foo'] = 'bar'
    holder.attributes = {'key1': 'val', 'key2': 'otherval'}
    """

    @property
    def attributes(self):
        if not '_attributes' in self.__dict__:
            self._attributes = AttributeProxy(self)
        return self._attributes

    @attributes.setter
    def attributes(self, value):
        self.attributes.clear()
        self.attributes.update(value)

    @declared_attr
    def _discriminator(cls):
        return cls.__name__.lower()

    @declared_attr
    def _attribute_association_id(cls):
        return Column('attribute_association_id', Integer, ForeignKey(
            'attribute_associations.id'))

    @declared_attr
    def _attribute_association(cls):
        return relationship('AttributeAssociation', backref=backref(
            '%s_owner' % cls._discriminator, uselist=False), cascade=
            'all, delete')


class User(AttributeHolder, Deletable, Base):
    """
    A user.

    Has a unique id and username, as well as a password and zero or
    more YubiKeys. Each user also has a key-value mapping of attributes.

    @cvar id: ID column.
    @cvar name: Name column.
    @cvar auth: Authentication data column.
    @cvar yubikeys: Dict of the users YubiKeys, with public IDs as keys.
    @cvar attributes: Dict of the users attributes.
    """
    __tablename__ = 'users'

    id = Column(Integer, Sequence('user_id_seq'), primary_key=True)
    name = Column(String(32), nullable=False, unique=True)
    auth = Column(String(128))
    yubikeys = relationship(
        'YubiKey',
        secondary=user_yubikeys,
        backref='users',
        order_by='YubiKey.prefix',
        collection_class=attribute_mapped_collection('prefix'),
    )

    def __init__(self, name, password):
        self.name = name
        self.set_password(password)
        super(User, self).__init__()

    def assign_yubikey(self, prefix_or_otp):
        """
        Assigns a YubiKey to the user.

        @param prefix_or_otp: The public ID of a YubiKey, or a full OTP.
        @type prefix_or_otp: string
        """
        if len(prefix_or_otp) > 32:
            prefix = prefix_or_otp[:-32]
        else:
            prefix = prefix_or_otp

        session = Session.object_session(self)
        existing_key = session.query(YubiKey).filter(
            YubiKey.prefix == prefix).first()

        if existing_key:
            existing_key.users.append(self)
        else:
            self.yubikeys[prefix] = YubiKey(prefix)
        return self.yubikeys[prefix]

    def set_password(self, password):
        """
        Sets the password of the user. An empty String is treated as no
        password. If no password is set the user either can't log in (default)
        or, if the ALLOW_EMPTY_PASSWORDS setting is set to True, the user may
        log in without providing a password.

        @param password: The new password to set for the user.
        @type password: string
        """
        if password:
            self.auth = pwd_context.encrypt(password)
        else:
            self.auth = None

    def validate_password(self, password):
        """
        Validates a password.

        @param password: A password to validate for the user.
        @type password: string

        @return: True if the password was valid, False if not.
        @rtype bool
        """
        if not password:
            return settings['allow_empty'] is True

        valid, new_auth = pwd_context.\
            verify_and_update(password, self.auth)
        if valid:
            if new_auth:
                self.auth = new_auth
            return True
        return False

    def validate_otp(self, otp):
        """
        Validates a YubiKey OTP (One Time Password) for the user.

        @param otp: The OTP to validate.
        @type otp: string

        @return: True if the OTP was valid, and from a YubiKey belonging
        to the user. False if not.
        @rtype: bool
        """
        prefix = otp[:-32]
        otp_valid = validate_otp(otp)
        if prefix in self.yubikeys:
            return otp_valid and self.yubikeys[prefix].enabled

        return False

    @property
    def data(self):
        return {
            'id': self.id,
            'name': self.name,
            'attributes': self.attributes.copy(),
            'yubikeys': self.yubikeys.keys()
        }

    def __repr__(self):
        return (
            "User(id: %r, name: '%s', yubikeys: %r, attributes: %r)" %
            (self.id, self.name, self.yubikeys.keys(), self.attributes)
        ).encode('utf-8')


class YubiKey(AttributeHolder, Deletable, Base):
    """
    A reference to a YubiKey.

    This class connects a particular YubiKey to a C{User} in a many-to-one
    relationship. It is also used to validate OTPs (One Time Passwords)
    from the YubiKey.

    @cvar id: ID column.
    @cvar prefix: YubiKey public ID column.
    @cvar user_id: FK of the owner.
    """
    __tablename__ = 'yubikeys'

    id = Column(Integer, Sequence('yubikey_id_seq'), primary_key=True)
    prefix = Column(String(32), nullable=False, unique=True)
    enabled = Column(Boolean, default=True)

    def __init__(self, prefix):
        self.prefix = prefix
        super(YubiKey, self).__init__()

    @property
    def data(self):
        return {
            'id': self.id,
            'prefix': self.prefix,
            'enabled': self.enabled,
            'attributes': self.attributes.copy()
        }

    def __repr__(self):
        return ("YubiKey(id: %r, prefix: '%s')" %
                (self.id, self.prefix)).encode('utf-8')



#
# Copyright (c) 2013 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

__all__ = [
    'AttributeType',
    'PERMS'
]

from sqlalchemy import Sequence, Column, Boolean, Integer, String
from sqlalchemy.ext.declarative import declarative_base

import re

from yubiauth.util.model import Deletable

Base = declarative_base()

PERMS = {
    'ALL': 1,
    'USER': 2,
    'ADMIN': 3,
    'NOBODY': 100
}


def clamp_perms(view_perms, edit_perms):
    if view_perms and not view_perms in PERMS:
        raise ValueError("Invalid value for view_perms: '%d'" % view_perms)
    if edit_perms and not edit_perms in PERMS:
        raise ValueError("Invalid value for edit_perms: '%d'" % view_perms)
    if view_perms > edit_perms:
        if edit_perms:
            raise ValueError("view_perms > edit_perms")
        edit_perms = view_perms

    return view_perms, edit_perms


class AttributeType(Deletable, Base):
    __tablename__ = 'attribute_types'

    id = Column(Integer, Sequence('attribute_type_id_seq'), primary_key=True)
    name = Column(String(32), unique=True, nullable=False)
    pattern = Column(String(128), nullable=False, default='.*')
    required = Column(Boolean, default=False)
    view_perms = Column(Integer, default=PERMS['ADMIN'])
    edit_perms = Column(Integer, default=PERMS['USER'])

    def __init__(self, name, pattern=None, required=False, view_perms=0,
                 edit_perms=0):
        view_perms, edit_perms = clamp_perms(view_perms, edit_perms)

        self.type = type
        self.name = name
        if pattern:
            self.pattern = pattern
        self.required = required
        if view_perms:
            self.view_perms = view_perms
        if edit_perms:
            self.edit_perms = edit_perms

    def validate(self, value):
        if not self.required and value is None:
            return True
        regex = re.compile(self.pattern)
        return bool(regex.match(value))

    @property
    def key(self):
        # TODO: normalize name
        return '__%s' % self.name

    @property
    def data(self):
        return {
            'name': self.name,
            'key': self.key,
            'pattern': self.pattern,
            'required': self.required,
            'view_perms': self.view_perms,
            'edit_perms': self.edit_perms
        }

    def __repr__(self):
        return (
            "AttributeType(name: '%s', key: '%s', pattern: '%s')" %
            (self.name, self.key, self.pattern)
        ).encode('utf-8')



#
# Copyright (c) 2013 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

import os
from wsgiref.simple_server import make_server
from webob import exc
from webob.dec import wsgify
import logging

from yubiauth.core.rest import application as core_rest
from yubiauth.client.rest import application as client_rest
from yubiauth.client.web import application as client_web
from yubiauth.util.static import DirectoryApp, FileApp

STATIC_ASSETS = ['js', 'css', 'img', 'favicon.ico']


class YubiAuthAPI(object):
    def __init__(self):
        base_dir = os.path.dirname(__file__)
        static_dir = os.path.join(base_dir, 'static')
        static_app = DirectoryApp(static_dir)
        favicon_app = FileApp(os.path.join(static_dir, 'favicon.ico'))

        self._apps = {
            'core': core_rest,
            'client': client_rest,
            'ui': client_web,
            'static': static_app,
            'favicon.ico': favicon_app
        }

    @wsgify
    def __call__(self, request):
        base_path = request.environ.get('BASE_PATH', '/')
        if not request.script_name and request.path_info.startswith(base_path):
            request.script_name = base_path
            request.path_info = request.path_info[len(base_path):]

        app_key = request.path_info_pop()
        if app_key in self._apps:
            return request.get_response(self._apps[app_key])

        raise exc.HTTPNotFound


application = YubiAuthAPI()

if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)

    httpd = make_server('localhost', 8080, application)
    httpd.serve_forever()



#
# Core API settings
#

# Database configuration string
DATABASE_CONFIGURATION = 'sqlite:///:memory:'

# YubiKey Validation Server URLs
YKVAL_SERVERS = [
    'https://api.yubico.com/wsapi/2.0/verify',
    'https://api2.yubico.com/wsapi/2.0/verify',
    'https://api3.yubico.com/wsapi/2.0/verify',
    'https://api4.yubico.com/wsapi/2.0/verify',
    'https://api5.yubico.com/wsapi/2.0/verify'
]

# YubiKey Validation server API key and secret
YKVAL_CLIENT_ID = 11004
YKVAL_CLIENT_SECRET = '5Vm3Zp2mUTQHMo1DeG9tdojpc1Y='

# Allow users with no password to log in without providing a password.
# When set to False (default), a user with no password will be unable to log
# in.
#
# WARNING: In combination with the SECURITY_LEVEL settings, this setting may
# allow a user to log in using ONLY a username. If you're unsure if you need
# this, leave it to False.
ALLOW_EMPTY_PASSWORDS = False

# Use a YubiHSM for increased security
# This requires the pyhsm package, and should be used with the yhsm-daemon
# utility that comes with it.
USE_HSM = False

# YubiHSM, only used if USE_HSM is True.
# Setting the 'YHSM_DEVICE' environment variable will override this.
YHSM_DEVICE = 'yhsm://localhost:5348'

# Passlib configuration
# Will default to using yhsm_pbkdf2_sha1 for password hashes if a HSM is
# available.
CRYPT_CONTEXT = {
    'schemes': ['yhsm_pbkdf2_sha1', 'sha256_crypt'] if USE_HSM
    else ['sha256_crypt'],
    'deprecated': ['sha256_crypt'] if USE_HSM else [],
    'default': 'yhsm_pbkdf2_sha1' if USE_HSM else 'sha256_crypt',
    'yhsm_pbkdf2_sha1__key_handle': 1,
    'all__vary_rounds': 0.1,
    'sha256_crypt__min_rounds': 80000,
    'admin__sha256_crypt__min_rounds': 160000
}

#
# Client API settings
#

# Defines who is required to provide a YubiKey OTP when logging in.
# The available levels are:
# 0 = Permissive. OTPs are not required to authenticate, by anyone.
#
# 1 = Require when provisioned. OTPs are required by all users that have a
# YubiKey assigned to them.
#
# 2 = Always require. OTPs are required by all users. If no YubiKey has been
# assigned, that user cannot log in, unless auto-provisioning is enabled.
#
# WARNING: In combination with the ALLOW_EMPTY_PASSWORDS settings, setting
# this to 0 may allow a user to log in using ONLY a username.
SECURITY_LEVEL = 1

# Auto-provision YubiKeys.
# When enabled, an attempt to authenticate a user that doesn't have a YubiKey
# assigned with a valid YubiKey OTP, will cause that YubiKey to become
# automatically assigned to the user.
AUTO_PROVISION = True

# When set to True, allow users authenticating with a YubiKey to omit their
# username, using the OTP to do a lookup on the username.
YUBIKEY_IDENTIFICATION = False

# When set to True, allow users to register accounts via the client API.
ENABLE_USER_REGISTRATION = False

# Beaker configuration, which is used for session management.
BEAKER = {
    'session.lock_dir': '/tmp/cache/lock',
    'session.type': 'ext:database',
    'session.auto': True,
    'session.key': 'YubiAuth-Session',
    'session.secret': 'LGnYUI5ZeSQ5ooGykySO9vMHKKyGwoC1',
}



#!/usr/bin/python
#
# Copyright (c) 2013 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

from setuptools import setup
import sys
import os
from release import release

tests_require = ['WebTest', 'mock']

# Don't load custom settings (for tests, etc.)
os.environ['YUBIAUTH_SETTINGS'] = '/dev/null'

# Require pyhsm if running hsm tests
if 'hsm' in sys.argv:
    tests_require.append('pyhsm')

setup(
    name='yubiauth',
    version='0.3.2',
    author='Dain Nilsson',
    author_email='dain@yubico.com',
    maintainer='Yubico Open Source Maintainers',
    maintainer_email='ossmaint@yubico.com',
    url='https://github.com/Yubico/yubiauth',
    license='BSD 2 clause',
    packages=['yubiauth', 'yubiauth.core', 'yubiauth.client', 'yubiauth.util'],
    include_package_data=True,
    setup_requires=['nose>=1.0'],
    install_requires=['SQLAlchemy', 'WebOb', 'passlib', 'yubico-client',
                      'Beaker', 'Jinja2', 'WTForms'],
    test_suite="nose.collector",
    tests_require=tests_require,
    cmdclass={'release': release},
    classifiers=[
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Topic :: Security :: Cryptography',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Application',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)



# Copyright (c) 2007 Ian Bicking and Contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import mimetypes
import os

from webob import exc
from webob.dec import wsgify
from webob.response import Response

__all__ = [
    'FileApp', 'DirectoryApp',
]

mimetypes._winreg = None # do not load mimetypes from windows registry
mimetypes.add_type('text/javascript', '.js') # stdlib default is application/x-javascript
mimetypes.add_type('image/x-icon', '.ico') # not among defaults


class FileApp(object):
    """An application that will send the file at the given filename.

    Adds a mime type based on `mimetypes.guess_type()`.
    """

    def __init__(self, filename, **kw):
        self.filename = filename
        content_type, content_encoding = mimetypes.guess_type(filename)
        kw.setdefault('content_type', content_type)
        kw.setdefault('content_encoding', content_encoding)
        kw.setdefault('accept_ranges', 'bytes')
        self.kw = kw

    @wsgify
    def __call__(self, req):
        if req.method not in ('GET', 'HEAD'):
            return exc.HTTPMethodNotAllowed("You cannot %s a file" %
                                            req.method)
        try:
            stat = os.stat(self.filename)
        except (IOError, OSError) as e:
            msg = "Can't open %r: %s" % (self.filename, e)
            return exc.HTTPNotFound(comment=msg)

        try:
            file = open(self.filename, 'rb')
        except (IOError, OSError) as e:
            msg = "You are not permitted to view this file (%s)" % e
            return exc.HTTPForbidden(msg)

        return Response(
            app_iter = FileIter(file),
            content_length = stat.st_size,
            last_modified = stat.st_mtime,
            #@@ etag
            **self.kw
        ).conditional_response_app


class FileIter(object):
    def __init__(self, file):
        self.file = file

    def app_iter_range(self, seek=None, limit=None, block_size=1<<16):
        """Iter over the content of the file.

        You can set the `seek` parameter to read the file starting from a
        specific position.

        You can set the `limit` parameter to read the file up to specific
        position.

        Finally, you can change the number of bytes read at once by setting the
        `block_size` parameter.
        """

        if seek:
            self.file.seek(seek)
            if limit is not None:
                limit -= seek
        try:
            while True:
                data = self.file.read(min(block_size, limit)
                                      if limit is not None
                                      else block_size)
                if not data:
                    return
                yield data
                if limit is not None:
                    limit -= len(data)
                    if limit <= 0:
                        return
        finally:
            self.file.close()

    __iter__ = app_iter_range


class DirectoryApp(object):
    """An application that dispatches requests to corresponding `FileApp`s based
    on PATH_INFO.

    This app double-checks not to serve any files that are not in a
    subdirectory.  To customize `FileApp` instances creation, override
    `make_fileapp` method.
    """

    def __init__(self, path, **kw):
        self.path = os.path.abspath(path)
        if not self.path.endswith(os.path.sep):
            self.path += os.path.sep
        assert os.path.isdir(self.path)
        self.fileapp_kw = kw

    def make_fileapp(self, path):
        return FileApp(path, **self.fileapp_kw)

    @wsgify
    def __call__(self, req):
        path = os.path.abspath(os.path.join(self.path,
                                            req.path_info.lstrip('/')))
        if not os.path.isfile(path):
            return exc.HTTPNotFound(comment=path)
        elif not path.startswith(self.path):
            return exc.HTTPForbidden()
        else:
            return self.make_fileapp(path)



from nose import with_setup

from yubiauth import create_tables
from yubiauth.core import YubiAuth

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

engine = create_engine('sqlite:///:memory:', echo=True)
create_tables(engine)
Session = sessionmaker(bind=engine)
auth = None


def setup():
    global auth
    auth = YubiAuth(Session())
    teardown()
    auth.create_user('user1', 'p4ssw0rd')
    auth.create_user('user2', 'foo')


def teardown():
    for user in auth.query_users():
        auth.get_user(user['id']).delete()


@with_setup(setup, teardown)
def test_yubikey_assignment():
    user1 = auth.get_user('user1')
    user1.assign_yubikey('cccccccccccb')

    user2 = auth.get_user('user2')
    user2.assign_yubikey('cccccccccccd')
    user2.assign_yubikey('ccccccccccce')

    assert auth.commit()

    assert user1.yubikeys['cccccccccccb']
    assert len(user2.yubikeys) == 2



#
# Copyright (c) 2013 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

__all__ = [
    'settings'
]

import sys
import imp
import errno
import os
import default_settings


SETTINGS_FILE = os.getenv('YUBIAUTH_SETTINGS',
                          '/etc/yubico/auth/yubiauth.conf')

VALUES = {
    #Core
    'DATABASE_CONFIGURATION': 'db',
    'YKVAL_SERVERS': 'ykval',
    'YKVAL_CLIENT_ID': 'ykval_id',
    'YKVAL_CLIENT_SECRET': 'ykval_secret',
    'ALLOW_EMPTY_PASSWORDS': 'allow_empty',
    'USE_HSM': 'use_hsm',
    'YHSM_DEVICE': 'yhsm_device',
    'CRYPT_CONTEXT': 'crypt_context',
    #Client
    'CORE_URL': 'core_url',
    'SECURITY_LEVEL': 'security_level',
    'AUTO_PROVISION': 'auto_provision',
    'YUBIKEY_IDENTIFICATION': 'yubikey_id',
    'ENABLE_USER_REGISTRATION': 'registration',
    'BEAKER': 'beaker'
}


def parse(conf, settings={}):
    for confkey, settingskey in VALUES.items():
        try:
            settings[settingskey] = conf.__getattribute__(confkey)
        except AttributeError:
            pass
    return settings


settings = parse(default_settings)

dont_write_bytecode = sys.dont_write_bytecode
try:
    sys.dont_write_bytecode = True
    user_settings = imp.load_source('user_settings', SETTINGS_FILE)
    settings = parse(user_settings, settings)
except IOError, e:
    if not e.errno in [errno.ENOENT, errno.EACCES]:
        raise e
finally:
    sys.dont_write_bytecode = dont_write_bytecode

if not 'session.url' in settings['beaker']:
    settings['beaker']['session.url'] = settings['db']

if not 'YHSM_DEVICE' in os.environ and 'yhsm_device' in settings:
    #The environment variable is the one that is actually used.
    os.environ['YHSM_DEVICE'] = settings['yhsm_device']



from webtest import TestApp
from yubiauth.client.rest import application, SESSION_COOKIE, SESSION_HEADER
from yubiauth.util.model import engine
from yubiauth import create_tables, YubiAuth
from utils import setting
from mock import patch


create_tables(engine)

app = TestApp(application)


def setup():
    auth = YubiAuth()
    for user in auth.query_users():
        auth.get_user(user['id']).delete()
    auth.commit()
    auth.create_user('user1', 'pass1')
    auth.create_user('user2', 'pass2')
    auth.commit()
    del auth


def test_login_logout():
    app.post(
        '/login',
        {'username': 'user1', 'password': 'pass1'}
    )

    # Uses the cookie set by the previous call
    username = app.get('/status').json['username']
    assert username == 'user1'

    app.get('/logout')
    app.get('/status', status=400)


def test_header_authentication():
    app.post(
        '/login',
        {'username': 'user1', 'password': 'pass1'}
    )

    session_id = app.cookies[SESSION_COOKIE]
    #Clear the cookie
    app.reset()
    app.get('/status', status=400)

    headers = {SESSION_HEADER: session_id}

    username = app.get('/status', headers=headers).json['username']
    assert username == 'user1'

    app.reset()
    app.get('/logout', headers=headers)
    app.reset()
    app.get('/status', headers=headers, status=400)


def test_update_password():
    app.post(
        '/login',
        {'username': 'user1', 'password': 'pass1'}
    )

    app.post('/password', {
        'oldpass': 'pass1', 'newpass': 'foobar'})

    assert app.post('/authenticate',
                    {'username': 'user1', 'password': 'foobar'}).json

    app.post('/password', {
        'oldpass': 'pass1', 'newpass': 'foobar'}, status=400)

    app.post('/password', {
        'oldpass': 'foobar', 'newpass': 'pass1'})
    app.get('/logout')


def test_empty_password_login():
    app.post(
        '/login',
        {'username': 'user1', 'password': 'pass1'}
    )

    app.post('/password', {
        'oldpass': 'pass1', 'newpass': ''})

    assert not app.post('/authenticate',
                        {'username': 'user1'}, status=400).json

    with setting(allow_empty=True):
        assert app.post('/authenticate',
                        {'username': 'user1'}).json

        app.post('/password', {
            'oldpass': '', 'newpass': 'pass1'})

    app.get('/logout')


@patch('yubiauth.util.utils.yubico', return_value=True)
def test_authentication_without_username(mock):
    app.post(
        '/login',
        {'username': 'user1', 'password': 'pass1'}
    )

    otp = 'c' * 44
    app.post('/yubikey', {'yubikey': otp, 'password': 'pass1'})

    assert not app.post('/authenticate',
                        {'otp': otp, 'password': 'pass1'},
                        status=400).json

    with setting(yubikey_id=True):
        assert app.post('/authenticate',
                        {'otp': otp, 'password': 'pass1'}).json

        assert not app.post('/authenticate',
                            {'otp': otp, 'password': 'wrongpass'},
                            status=400).json
        otp = 'd' * 44

        assert not app.post('/authenticate',
                            {'otp': otp, 'password': 'pass1'},
                            status=400).json

    app.get('/logout')


@patch('yubiauth.util.utils.yubico', return_value=True)
def test_single_factor_login(mock):
    otp = 'c' * 44
    with setting(yubikey_id=True, allow_empty=True):
        assert app.post('/authenticate', {'otp': otp}).json
        app.post('/login', {'otp': otp})
        status = app.get('/status').json
        assert status['username'] == 'user1'
        app.get('/logout')

    assert not app.post('/authenticate', {'otp': otp},
                        status=400).json
